import bluesky.callbacks


class BulkLiveTable(bluesky.callbacks.LiveTable):
    def bulk_events(self, keyed_docs):
        for uid, docs in keyed_docs.items():
            for doc in docs:
                self.event(doc)



import logging
from collections import OrderedDict
from IPython import get_ipython

from .fly1d import Fly1D
from .fly2d import Fly2D

logger = logging.getLogger(__name__)

ip = get_ipython()

try:
    ssx = ip.user_ns['ssx']
    ssy = ip.user_ns['ssy']
    ssz = ip.user_ns['ssz']
except KeyError as ex:
    logger.error('Ensure that fly scan motor "%s" is still in the startup '
                 'configuration with the same name.', ex,
                 exc_info=ex)
    raise

coord_sys = 5
prog_num = 50
axes = OrderedDict([('x', dict(positioner=ssx, axis_number=3)),
                    ('y', dict(positioner=ssy, axis_number=4)),
                    ('z', dict(positioner=ssz, axis_number=5))]
                    )
# change return_speed from 5.0 to 40 (H. Yan, 01/10/17)
# Reverted to 2ms dead time (Merlin requires at least 1.6ms)
configure_defaults = dict(return_speed=40.0,
                          dead_time=0.002,
                          fly_type='soft',
                          max_points=16384
                          )


class Fly1D_MLL(Fly1D):
    def __init__(self, cmd='fly1d', **kwargs):
        super().__init__(axes=axes, coord_sys=coord_sys, prog_num=prog_num,
                         cmd=cmd, configure_defaults=configure_defaults,
                         **kwargs)


class Fly2D_MLL(Fly2D):
    def __init__(self, cmd='fly2d', **kwargs):
        super().__init__(axes=axes, coord_sys=coord_sys, prog_num=prog_num,
                         cmd=cmd, configure_defaults=configure_defaults,
                         **kwargs)



import os
import logging

MODULE_PATH = os.path.dirname(os.path.abspath(__file__))
SCRIPT_PATH = os.path.join(MODULE_PATH, 'scripts')

logger = logging.getLogger(__name__)

from .log import log_setup

log_setup(logging.DEBUG)

# from .fly1d import Fly1D
# from .fly2d import Fly2D
#
# from .fly_mll import (Fly1D_MLL, Fly2D_MLL)
# from .fly_zp import (Fly1D_ZP, Fly2D_ZP)



import asyncio
import logging
import functools
import time
import threading
from collections import OrderedDict

import numpy as np
# from bluesky.callbacks import CallbackBase
from bluesky.callbacks.mpl_plotting import QtAwareCallback

loop = asyncio.get_event_loop()
logger = logging.getLogger(__name__)


def catch_exceptions(fcn):
    @functools.wraps(fcn)
    def wrapped(*args, **kwargs):
        try:
            return fcn(*args, **kwargs)
        except Exception as ex:
            logger.error('Function call failed: %s(%s, %s)', fcn, args, kwargs,
                         exc_info=ex)
            raise

    return wrapped


class SubscanChecker:
    def __init__(self, subscan_rate):
        self._subscan_index = 0
        self._fut_subscan = None
        self._subscan_rate = subscan_rate

    def start_subscan_checker(self):
        if self._fut_subscan is None:
            self._fut_subscan = loop.call_soon(self.check_subscan)

    def stop_subscan_checker(self):
        if self._fut_subscan is None:
            return

        logger.debug('Cancelling subscan checker')
        try:
            self._fut_subscan.cancel()
        except Exception:
            pass
        finally:
            self._fut_subscan = None

    def check_subscan(self):
        flyer = self.flyer
        if flyer is not None and flyer.scan_count == 1:
            self._fut_subscan = None
            return

        try:
            if flyer is None or flyer.scan_count <= 1:
                pass
            else:
                last_index = self._subscan_index
                if flyer.scan_index > last_index:
                    self._subscan_index = flyer.scan_index
                    self.subscan_start(index=self._subscan_index)
        finally:
            self._fut_subscan = loop.call_later(self._subscan_rate,
                                                self.check_subscan)


class FlyDataCallbacks(SubscanChecker, QtAwareCallback):
    def __init__(self, flyer=None, subscan_rate=0.5):
        SubscanChecker.__init__(self, subscan_rate=subscan_rate)
        QtAwareCallback.__init__(self)

        self.flyer = flyer
        self._run_header = None

    @catch_exceptions
    def start(self, doc):
        self._run_header = doc
        logger.debug('Run header: scan_id=%d (%s)', doc['scan_id'],
                     self.__class__.__name__)

        self.scan_id = doc['scan_id']
        self.fast_axis = doc['fast_axis']
        dims = doc['shape']
        ndim = len(dims)
        self.num = np.product(dims)
        self._subscan_index = 0

        if ndim == 1:
            pass
        elif ndim == 2:
            self.start_subscan_checker()
        else:
            logger.error('Unsupported scan dimensions')
            return

        logger.debug('Number of points=%d', self.num)
        self.scan_started(doc, ndim, fast_axis=self.fast_axis,
                          **doc['plan_args'])

    @catch_exceptions
    def stop(self, doc):
        if self.flyer is None:
            logger.error('Flyer unset - not running scan_finished')
            return

        flyer = self.flyer
        self.stop_subscan_checker()

        while not flyer.done:
            logger.debug('Waiting for flyscan to finish')
            time.sleep(0.2)

        scan_data = self.flyer.data.get(self.scan_id, {})

        finish_func = (self.scan_failed if flyer.failed
                       else self.scan_finished)

        try:
            finish_func(doc, scan_data, cancelled=self.flyer.cancelled)
        finally:
            self._run_header = None
            self.flyer = None

    def scan_started(self, doc, ndim, **scan_kwargs):
        '''Re-implement me'''
        pass

    def scan_failed(self, doc, scan_data, **kwargs):
        '''Re-implement me'''
        pass

    def scan_finished(self, doc, scan_data, cancelled=False, **kwargs):
        '''Re-implement me'''
        pass

    def subscan_start(self, index=0, **kwargs):
        '''Re-implement me'''
        pass


def default_data_func(*arrays):
    return np.sum(arrays, axis=0)


class SignalDataHandler:
    def __init__(self, groups, data_func=None, point_signal=None):
        self.groups = groups
        self.point_signal = point_signal
        self.data_func = data_func
        self.signals = sum((signals for group, signals in groups.items()), [])
        self._updated = True
        self._lock = threading.RLock()
        self._calculated = None

        if point_signal is not None:
            point_signal.subscribe(self._value_updated)
        else:
            for sig in self.signals:
                sig.subscribe(self._value_updated)
        if not groups:
            raise ValueError('Must have at least one signal')

    @property
    def updated(self):
        '''Any signals changed since the last calculation?'''
        return self._updated

    def _value_updated(self, **kwargs):
        with self._lock:
            self._updated = True

    def get_group_data(self, key, npts):
        group_data = []
        for signal in self.groups[key]:
            try:
                value = signal.get(count=npts)
            except Exception as ex:
                logger.debug('Caget failed for signal %s', signal,
                             exc_info=ex)
                value = None

            if value is not None:
                try:
                    len(value)
                except TypeError:
                    value = [value]

                if len(value) < npts:
                    value = np.concatenate((value, [0] * (npts - len(value))))
                group_data.append(value)
            else:
                group_data.append(np.zeros(npts))
        return group_data

    @catch_exceptions
    def calc_data(self, npts):
        with self._lock:
            if not self._updated:
                return self._calculated
            elif npts <= 0:
                return OrderedDict((label, [])
                                   for label, signals in self.groups.items())

            data = OrderedDict()
            data_func = (self.data_func if self.data_func is not None
                         else default_data_func)

            for label, signals in self.groups.items():
                group_data = self.get_group_data(label, npts)
                data[label] = data_func(*group_data)

            self._updated = False
            self._calculated = data

        return OrderedDict(self._calculated)

    @property
    def num_points(self):
        if self.point_signal is None:
            raise ValueError('No point signal set')

        return self.point_signal.get(use_monitor=False)


class SubscanDataHandler:
    def __init__(self, key_groups, data_func=None):
        self.key_groups = key_groups
        self.data_func = data_func

    def calc_data(self, flyer, scan_id):
        try:
            data = flyer.data[scan_id]
        except KeyError:
            return OrderedDict()

        data_func = (self.data_func if self.data_func is not None
                     else default_data_func)
        return OrderedDict((main_key, data_func(*[data[key] for key in keys]))
                           for main_key, keys in self.key_groups.items())



import asyncio
import threading
import logging
import logging.handlers
import numpy as np
import os
import time
import tqdm

from datetime import datetime
from itertools import zip_longest
from collections import OrderedDict
import pandas as pd

import IPython

from ophyd import Device
from ophyd.status import Status
from ophyd.device import Staged
from ophyd.areadetector import FilePlugin

from nslsii.detectors.xspress3 import Xspress3Detector
from hxntools.shutter import (shutter_open, shutter_close)

from . import SCRIPT_PATH
from ppmac.pp_comm import (GPError, ScriptCancelled,
                           TimeoutError as PpmacTimeoutError)
from .ppmac_util import (ppmac_connect, filter_ppmac_data as filter_data,
                         gather_setup, is_closed_loop)
from .log import log_time

# from .plot import plot_position

# ip = IPython.get_ipython()
# global_state = ip.user_ns['gs']
# zebra = ip.user_ns['zebra']
# sclr1 = ip.user_ns['sclr1']
# from hxntools import HxnTriggeringScaler
# sclr1 = HxnTriggeringScaler('XF:03IDC-ES{Sclr:1}', name='sclr1')

logger = logging.getLogger(__name__)
UNSTAGE_TIME_LIMIT = 300


class FlyStatus(Status):
    pass


def get_dataframe(df_data, n_points):
    '''Create a dataframe from df_data, adjusting the size of lists if
    necessary

    Note: this should not really be happening (and hasn't recently I believe),
    but we'd rather a point or two to be clipped off than for the whole scan to
    die in case something went a bit wrong.
    '''
    for key in df_data.keys():
        len_ = len(df_data[key])
        if len_ < n_points:
            df_data[key] = np.pad(df_data[key], (0, n_points - len_),
                                  mode='constant')
        elif len_ > n_points:
            df_data[key] = df_data[key][:n_points]

    return pd.DataFrame(df_data)


def get_scaler_info(scaler, channels, *, with_data=False, num_points=None):
    '''Get scaler information, optionally including data

    Parameters
    ----------
    channels : list
        List of channels
    with_data : bool, optional
        Include MCA data
        For channel 1, this will be converted to seconds
    num_points : int, optional
        Number of points to get from the data PV, default (None) will get all
        available
    '''
    for chan in channels:
        key = '{}_ch{}'.format(scaler.name, chan)
        mca = scaler.mca_by_index[chan]

        if with_data:
            data = mca.spectrum.get(count=num_points, as_numpy=True)

        if chan == 1:
            # Channel 1 is special, as we know it counts the alive time of
            # the scaler:
            key = 'scaler_alive'
            if with_data:
                yield key, mca, data.astype(float) / 50e3
            else:
                yield key, mca
        elif with_data:
            yield key, mca, data.astype(int)
        else:
            yield key, mca


def dataframe_to_bluesky_data(df, *, timestamp_key='timestamp'):
    '''Take pandas dataframe -> generator of dictionaries for bluesky collect

    Parameters
    ----------
    df : pd.DataFrame
    '''
    keys = list(df.keys())
    keys.remove(timestamp_key)
    frame_timestamps = list(df[timestamp_key])
    keyed_timestamps = ({k: ts for k in keys}
                        for ts in frame_timestamps)

    v = [df[k] for k in keys]
    data = ({k: val for k, val in zip(keys, row)} for row in
            zip_longest(*v, fillvalue=np.nan))

    return ({'time': ts,
             'timestamps': keyed_ts,
             'data': _data,
             }
            for _data, ts, keyed_ts in zip(data, frame_timestamps,
                                           keyed_timestamps)
            )


class FlyBase(Device):
    '''FlyScan base class

    NOTE: this is only made a Device for bluesky compatibility currently(?)
    '''

    _cancel_script = 'cancel.txt'

    def __init__(self, cmd, ppmac, configure_defaults=None):
        super().__init__('', name=self.__class__.__name__, parent=None)

        self.stream_name = 'primary'
        self._sync_variable = 'Q1'
        self._status_variable = 'Q2'
        self._progress_bar = None
        self._collect_data = []
        self._data_desc = OrderedDict()
        self._cmd = cmd  # scan instance name user sees
        self._metadata = {}
        self._script_running = False
        self._failed = False
        self._start_positions = {}
        # TODO 1D scan is limited by buffer size, but 2D is not
        self._raw_gather_data = bytearray()
        self._debug_data = {}
        self._kickoff_status = None
        self._collect_status = None
        self._scan_index = 1
        self.scan_id = None
        self._total_points = None
        self.scan_command = ''
        self.estimated_time = 0.0

        ipython = IPython.get_ipython()
        self.zebra = ipython.user_ns['zebra']
        self.scaler = ipython.user_ns['sclr1']

        if configure_defaults is None:
            configure_defaults = {}

        self.configure_defaults = configure_defaults

        if ppmac is None:
            ppmac = ppmac_connect()

        self.ppmac = ppmac
        self.gpascii = ppmac.gpascii

    @property
    def scan_count(self):
        return len(self._subscans)

    @property
    def done(self):
        try:
            return self._collect_status.done
        except AttributeError:
            pass

    @property
    def total_points(self):
        return self._total_points

    @property
    def scan_index(self):
        return self._scan_index

    @property
    def failed(self):
        return self._failed

    @property
    def cancelled(self):
        return self._cancelled.is_set()

    def configure(self, *args, **kwargs):
        '''Configure the fly scan

        * Sets up metadata
        * Creates a 'scan command' based on args/kwargs to this function
        * Calls _configure (which is based on the fly-scan type) that populates
          the sub-scan list
        '''
        # assemble the scan_command
        arg_str = [arg.name if hasattr(arg, 'name') else repr(arg)
                   for arg in args]
        kwarg_str = ['{}={!r}'.format(k, v) for k, v in kwargs.items()]

        self.scan_command = '{}({})'.format(self._cmd,
                                            ', '.join(arg_str + kwarg_str))

        self._metadata['scan_command'] = self.scan_command

        # record the starting positions of the motors
        self._start_positions = self.positions

        # configure all of the subscans
        subscans = self._configure(*args, **kwargs)

        if subscans:
            scan_info = subscans[0]
            macros = scan_info['macros']
            self._metadata['script_path'] = scan_info['script_fn']
            self._metadata['fly_type'] = scan_info['fly_type']
            self._metadata['axes'] = macros['axis_names']

        self._collect_desc = self._get_collect_description()
        return kwargs, kwargs

    def kickoff(self, *, subscan=0, name=None):
        '''Bluesky kickoff to start a specific subscan
        '''
        self._kickoff_status = FlyStatus()
        self._collect_status = FlyStatus()
        # self.stream_name = name

        def run_fly():
            self._kickoff_status._finished(success=True)

            try:
                self._collect_data = self.run_subscan(subscan)
            except Exception:
                self._failed = True
                self._collect_status._finished(success=False)
                raise
            finally:
                self._collect_status._finished(success=True)

        loop = asyncio.get_event_loop()
        loop.run_in_executor(None, run_fly)
        return self._kickoff_status

    def _get_collect_description(self):
        self._roi_keys = []
        for xsp in self.detectors_of_type(Xspress3Detector):
            for desc in xsp.describe_collect():
                self._data_desc.update(**desc)
            roi_keys = list(roi.name for roi in xsp.enabled_rois)
            self._roi_keys.extend(roi_keys)

        return {self.stream_name: self._data_desc}

    def describe_collect(self):
        if self._collect_desc is None:
            self._collect_desc = self._get_collect_description()

        return self._collect_desc

    def complete(self):
        '''Return status object for data collection status'''
        return self._collect_status

    def collect(self):
        '''The collected data from run_subscan'''
        yield from self._collect_data

    def pause(self):
        if self._collect_status.done and not self._script_running:
            return

        logger.info('Pause requested')

        if self.cancelled:
            logger.debug('Pause called twice')
            return

        self._cancelled.set()

        if self._script_running:
            logger.warning('Stopping the scan motion program...')
        elif self.failed:
            return
        else:
            logger.info('Scan not running - waiting for data to be '
                        'written')

        try:
            logger.warning('Waiting for motion program to exit safely')
            while not self._collect_status.done:
                time.sleep(0.1)
        except KeyboardInterrupt:
            logger.warning('Scan interrupted; data may not have been saved')

    def resume(self):
        if self.cancelled:
            raise RuntimeError('Cannot resume a hard-paused flyscan')

    @property
    def md(self):
        '''scan metadata for runstart header'''
        return self._metadata

    def _setup_axes(self, axes):
        '''
        Parameters
        ----------
        axes : OrderedDict
            {simple_axis_name: {'positioner': ophyd.Positioner,
                                'axis': axis_number}}
        '''
        # axes: name -> number
        self.axes = OrderedDict()

        # axis_to_positioner: name -> positioner
        self.axis_to_positioner = OrderedDict()

        # positioner_to_axis: positioner -> simple axis name for motion script
        self.positioner_to_axis = OrderedDict()
        for axis_name, axis_info in axes.items():
            axis_num = axis_info['axis_number']
            pos = axis_info['positioner']

            self.axes[axis_name] = axis_num
            if not is_closed_loop(self.gpascii, axis_num):
                raise RuntimeError('Axis {} not in closed loop'
                                   ''.format(axis_name))

            self.axis_to_positioner[axis_name] = pos
            self.positioner_to_axis[pos] = axis_name

        self.number_to_axis = OrderedDict((num, axis)
                                          for axis, num in self.axes.items())
        self.axis_keys = [positioner.name
                          for axis_name, positioner in
                          self.axis_to_positioner.items()]

    @property
    def positions(self):
        '''Positions for all axes {axis_simple_name: position}'''
        home_positions = self.home_positions
        return {axis: (self.gpascii.get_variable('Motor[%d].ActPos' % num,
                                                 type_=float) -
                       home_positions[axis])
                for num, axis in self.number_to_axis.items()}

    @property
    def home_positions(self):
        '''Home position for all axes {axis_simple_name: home_position}'''
        return {axis: self.gpascii.get_variable('Motor[%d].HomePos' % num,
                                                type_=float)
                for num, axis in self.number_to_axis.items()}

    @property
    def motor_numbers(self):
        '''motor numbers for the scan: 3, 4, and/or 5'''
        for name in self.simple_motor_names:
            yield self.axes[name]

    @property
    def simple_motor_names(self):
        '''motor names for the scan: x, y, and/or z'''
        for pos in self.scanning_positioners:
            yield self._get_axis_name(pos)

    def _move_all_axes(self, motors_targets, *, wait=True, msg='moved to'):
        '''Move motors to the starting position

        Parameters
        ----------
        motors_targets : list
            (axis_name, target)
        '''
        def _get_motor(motor):
            if motor in self.axes:
                return self.axes[motor]
            return motor

        motors_targets = [(_get_motor(motor), target)
                          for motor, target in motors_targets]

        for motor, target in motors_targets:
            self.gpascii.jog(motor, target, relative=False, wait=False)

        if wait:
            for motor, target in motors_targets:
                try:
                    self.gpascii.jog(motor, target, relative=False, wait=True)
                except PpmacTimeoutError:
                    logger.warning('Failed to move motor %s into position',
                                   motor)
                else:
                    logger.debug('Motor %s %s: %g', motor, msg, target)

    def _get_axis_name(self, motor):
        '''simple axis name of motor, such as "x", 'y", or "z"'''
        if motor in self.axes:
            return motor

        try:
            return self.positioner_to_axis[motor]
        except KeyError:
            raise KeyError('Unknown motor for this type of fly scan')


class FlyScan(FlyBase):

    def __init__(self, ppmac=None, script_path=SCRIPT_PATH, detectors=None,
                 shutter_control=False, coord_sys=5,
                 prog_num=50, axes=None, cmd='fly',
                 trigger_var='m100', scaler_channels=None,
                 configure_defaults=None):

        super().__init__(cmd=cmd, ppmac=ppmac,
                         configure_defaults=configure_defaults)

        self.coord_sys = coord_sys
        self.prog_num = prog_num

        self._setup_axes(axes)

        if detectors is None:
            detectors = []

        self.data = {}
        self.gather_client = self.ppmac.fast_gather
        self._script_path = script_path
        self.detectors = list(detectors)

        if self.zebra not in self.detectors:
            self.detectors.append(self.zebra)

        if self.scaler not in self.detectors:
            self.detectors.append(self.scaler)

        # self._debug = {'target_pos': True}
        self._debug = {}
        self._cancelled = threading.Event()
        self._shutter_control = shutter_control
        self._times = {}
        self._gate_enable = 'high_to_low'
        self._trigger_var = trigger_var
        self._roi_keys = []
        self.detector_status = []

        if scaler_channels is None:
            scaler_channels = [1]

        self._scaler_channels = list(sorted(scaler_channels))
        if 1 not in self._scaler_channels:
            self._scaler_channels.insert(0, 1)

        self._bulk_readable = []
        for det in detectors:
            for attr in det._sub_devices:
                sub_dev = getattr(det, attr)
                if hasattr(sub_dev, 'bulk_read'):
                    self._bulk_readable.append(sub_dev)

        logger.debug('Recording scaler channels: %s', self._scaler_channels)
        logger.debug('Bulk-readable detectors: %s',
                     ', '.join(sub_dev.name for sub_dev in self._bulk_readable)
                     )

    def stop(self):
        logger.info('Flyscan stop requested')
        if self.cancelled:
            logger.debug('Stop called twice')
            return

        if self._collect_status.done and not self._script_running:
            return

        self._cancelled.set()

        if self._script_running:
            logger.warning('Stopping the scan motion program...')
        elif self.failed:
            return
        else:
            logger.info('Scan not running - waiting for data to be '
                        'written')

    def _new_scan(self, dimensions=None, num_points=0, positioners=None,
                  paths=None, exposure_time=None, fly_type='soft',
                  **kwargs):
        self._cancelled.clear()
        self._failed = False
        self.dimensions = dimensions
        self.num_points = int(num_points)
        self.scanning_positioners = positioners
        self.paths = paths
        self.fly_type = fly_type
        self.exposure_time = exposure_time

        self._times['scan_init'] = time.time()

        self._detector_setup()

        self._times['detectors_configured'] = time.time()

        gather_setup(axis_numbers=list(self.axes.values()),
                     period=2,  # TODO can probably increase
                     gpascii=self.gpascii,
                     send_to_controller=True)

        # Ensure the 'capture buffer full' variable isn't set
        self.gpascii.set_variable(self._sync_variable, 0)

        if self._shutter_control:
            logger.debug('Opening shutter')
            shutter_open()

    def detectors_of_type(self, type_):
        '''Get all detectors which are of a specific (class) type'''
        for det in self.detectors:
            if isinstance(det, type_):
                yield det

    def _get_roi_data(self):
        rois = OrderedDict()

        for det in self.detectors_of_type(Xspress3Detector):
            try:
                logger.debug('%s HDF5 file: %s', det.name, det.hdf5_filename)
                rois.update(list(det.fly_collect_rois()))
                # rois is a list of (name, roi_info) pairs
            except Exception as ex:
                logger.error('Xspress3 failed to aggregate ROIs: %s.'
                             'If the HDF5 file was written properly, ROIs '
                             'can be calculated after the fact either '
                             'manually or with another tool such as pyxrf.',
                             ex, exc_info=ex)

        class NoROIData:
            ev_low = 0
            ev_high = 0

        # add missing keys, if necessary
        for key in self._roi_keys:
            if key not in rois:
                logger.error('ROI missing from detector readout: %s.  It will '
                             'be zero in the output, but can be re-calculated '
                             'using pyxrf', key)
                data = NoROIData()
                data.value = []
                rois[key] = data

        # remove unexpected keys, if necessary
        read_keys = set(rois.keys())
        expected_keys = set(self._roi_keys)
        for unexpected_key in read_keys.difference(expected_keys):
            logger.error('Unexpected ROI: %s', unexpected_key)
            del rois[unexpected_key]

        for key, roi in rois.items():
            yield key, roi

    @property
    def scan_id_string(self):
        if self.scan_count > 1:
            return '#{} (sub-scan {} of {})'.format(self.scan_id,
                                                    self.scan_index,
                                                    self.scan_count)
        return '#{}'.format(self.scan_id)

    def _process_data(self):
        self._times['data_aggregated'] = time.time()

        df_data = OrderedDict()

        # roi_info is a list of RoiTuples, with ev info and the data
        roi_info = list(self._get_roi_data())

        # and add the roi data itself
        for key, roi in roi_info:
            if key in df_data:
                logger.warning('Duplicate roi name %s; will not be stored',
                               key)
                continue

            df_data[key] = roi.value

        # tag on all filtered, gathered data
        gather_cols = self._get_gather_data()
        ppmac_data, self._debug_data = filter_data(
            servo_period=self.gpascii.servo_period,
            servo_t=gather_cols[0],
            enable=gather_cols[1],
            motor_cols=gather_cols[2:2 + len(self.axes)],
            other_cols=gather_cols[2 + len(self.axes):],
            axis_names=list(self.axes.keys()),
            axis_keys=self.axis_keys,
            have_target_pos=self._debug.get('target_pos', False),
            gate_enable=self._gate_enable,
            home_pos=self.home_positions,
        )

        # add ev information for all rois
        self._debug_data['rois'] = {key: (roi.ev_low, roi.ev_high)
                                    for key, roi in roi_info}

        df_data.update(ppmac_data)

        self._times['ppmac_run_time'] = self._debug_data['servo_time'][-1]

        scaler_info = get_scaler_info(self.scaler, self._scaler_channels,
                                      with_data=True,
                                      num_points=self.num_points)
        for key, mca, mca_data in scaler_info:
            df_data[key] = mca_data

        # TODO use EVR for all timestamps
        timestamps = (self._times['script_start'] +
                      np.asarray(df_data['elapsed_time']))
        df_data['timestamp'] = timestamps

        logger.debug('Total timestamps: %d', len(timestamps))

        # special-casing xspress3 for now :/
        for det in self.detectors_of_type(Xspress3Detector):
            # TODO: i think it will eventually move in this direction:
            det.flyer_timestamps.put(timestamps)
            df_data.update(det.bulk_read())
            logger.debug('Bulk read %d channels from %s',
                         len(det.channels), det.name)

        for readable in self._bulk_readable:
            df_data.update(readable.bulk_read(timestamps))

        self._times['data_filtered'] = time.time()

        self._record_times()

        try:
            new_df = get_dataframe(df_data, n_points=len(timestamps))
            full_df = pd.concat((self.data[self.scan_id], new_df))
            self.data[self.scan_id] = full_df
            return dataframe_to_bluesky_data(new_df, timestamp_key='timestamp')
        except Exception as ex:
            logger.error('Failed to aggregate into dataframe', exc_info=ex)
            raise

    def _record_times(self):
        if self._subscans:
            self._times['estimated_time'] = (self.estimated_time /
                                             len(self._subscans))
        else:
            self._times['estimated_time'] = self.estimated_time

        # setup time = time from initialization to the script starting
        log_time(logger, self._times, 'setup_time',
                 t1='scan_init', t2='script_start')

        # detector setup time is a portion of the overall setup time
        log_time(logger, self._times, 'detector_setup_time',
                 t1='scan_init', t2='detectors_configured')

        # aggregate_time = time when the script finishes to all data back
        log_time(logger, self._times, 'aggregate_time',
                 t1='script_finished', t2='data_aggregated')

        # filter_time = time to take the aggregated data and filter it
        log_time(logger, self._times, 'filter_time',
                 t1='data_aggregated', t2='data_filtered')

        # record just the ppmac script runtime
        log_time(logger, self._times, 'ppmac_run_time')

        # record the estimated script runtime
        log_time(logger, self._times, 'estimated_time')

    @property
    def table_columns(self):
        return [
            key
            for key, desc in self.describe_collect()[self.stream_name].items()
            if (key not in self._bulk_readable and desc['dtype'] != 'array')
        ]

    def _configure(self):
        for readable in self._bulk_readable:
            # bulk readable detectors have their own image data keys
            det = readable.parent
            self._data_desc[readable.image_name] = det.make_data_key()

        # motion axes and other data filtered from the gather buffer
        other_keys = ['elapsed_time', 'alive', 'dead', ]
        for key in list(self.axis_keys) + other_keys:
            self._data_desc[key] = dict(source='HW:ppmac_gather',
                                        shape=[],
                                        dtype='number')

        # all scaler-related signals
        def get_mca_desc(mca):
            spectrum_name = mca.spectrum.name
            desc = mca.spectrum.describe()
            return desc[spectrum_name]

        alive_desc = get_mca_desc(self.scaler.mca_by_index[1])
        self._data_desc['scaler_alive'] = alive_desc

        scaler_info = get_scaler_info(self.scaler, self._scaler_channels,
                                      with_data=False)
        for key, mca in scaler_info:
            self._data_desc[key] = get_mca_desc(mca)

    def trigger(self, on=False, off=False):
        if self._gate_enable == 'low_to_high':
            off_value, on_value = 0, 1
        elif self._gate_enable == 'high_to_low':
            off_value, on_value = 1, 0

        if on:
            logger.debug('Trigger on')
            self.gpascii.send_line('{}={}'.format(self._trigger_var, on_value))
            self.gpascii.sync()

        if off:
            logger.debug('Trigger off')
            self.gpascii.send_line('{}={}'.format(self._trigger_var,
                                                  off_value))
            self.gpascii.sync()

    def _detector_setup(self):
        self.trigger(off=True)
        logger.debug('Detector setup:')
        del self.detector_status[:]

        for det in self.detectors:
            settings = det.mode_settings
            settings.mode.put('external')
            settings.scan_type.put('fly')
            settings.total_points.put(self.num_points)

            logger.debug('Staging %s (settings: %s)', det.name,
                         settings.get())

            if det._staged == Staged.yes:
                logger.debug('Detector %s already staged - restaging',
                             det.name)
                det.unstage()

            det.stage()
            self.detector_status.append(det.trigger())

        for readable in self._bulk_readable:
            det = readable.parent
            # TODO: ensure image key is in the datum uids, because the detector
            # is being used outside of the normal bluesky methods...
            readable._datum_uids[readable.image_name] = []

    def _get_gather_data(self):
        n_samples, raw_data = self.gather_client.query_raw_data()
        self._raw_gather_data += raw_data

        t0 = time.time()
        dtypes = self.gather_client.query_types()
        parsed = self.gather_client._parse_raw_data(dtypes,
                                                    self._raw_gather_data)
        t1 = time.time()

        # TODO in the future, this can be done in parallel in a queue
        logger.debug('[sync mode] Raw data parsing took %f ms',
                     (1000. * (t1 - t0)))
        gathered, n_addr, n_samples = parsed
        cols = np.asarray(gathered)
        return np.asarray(cols)

    def unstage(self):
        for det in self.detectors:
            sub_devs = [getattr(det, sub_dev) for sub_dev in det._sub_devices]
            for sub_dev in sub_devs:
                if isinstance(sub_dev, FilePlugin):
                    t0 = time.time()
                    i = 0
                    while sub_dev.capture.get() != 0:
                        elapsed = time.time() - t0
                        if elapsed > UNSTAGE_TIME_LIMIT:
                            logger.error('%s unstage time limit reached. '
                                         'Stopping file plugin capture.',
                                         sub_dev.name,
                                         )
                            sub_dev.capture.put(0)
                            break
                        elif (i % 20) == 0:
                            logger.info('%s still acquiring: %d frames '
                                        '(%.1f sec until considered failure)',
                                        sub_dev.name,
                                        sub_dev.num_captured.get(),
                                        UNSTAGE_TIME_LIMIT - elapsed)
                        i += 1
                        time.sleep(0.1)
                    logger.info('%s acquired %d frames', sub_dev.name,
                                sub_dev.num_captured.get())

            logger.debug('Unstaging %s', det.name)
            det.unstage()

        for st in self.detector_status:
            logger.debug('Unstage detector status: %s', st)

    def _status_variable_updated(self, old, new):
        if self._progress_bar is None:
            return

        # status update
        try:
            new = int(new)
            old = int(old)
        except TypeError:
            return

        if new > old:
            try:
                self._progress_bar.update((new - old))
            except Exception:
                pass

    def _buffer_status_updated(self, old, new):
        # otherwise, Q1 for buffer status
        new = int(new)
        logger.debug('[sync status] %s=%d', self._sync_variable, new)
        if new == 1:
            logger.debug('[sync mode] sample count reached')
            t0 = time.time()

            _, raw_data = self.gather_client.query_raw_data()
            samples = self.gpascii.get_variable('Gather.Samples', type_=int)
            self.gpascii.set_variable(self._sync_variable, 0)

            t1 = time.time()
            logger.debug('[sync mode] line time %f ms (%d bytes, %d samples)',
                         (t1 - t0) * 1000., len(raw_data), samples)

            self._raw_gather_data += raw_data

    def _sync_callback(self, var, old, new):
        if var == self._status_variable:
            return self._status_variable_updated(old, new)
        elif var == self._sync_variable:
            return self._buffer_status_updated(old, new)

    def _run_and_wait(self, *args, num_points=None, **kwargs):
        '''Run the motion script for the already-configured subscan'''
        ipython = IPython.get_ipython()

        self.scan_id = ipython.user_ns['RE'].md['scan_id']
        if self.scan_id not in self.data:
            self.data[self.scan_id] = None

        logger.info("Scan ID: %s", self.scan_id_string)

        if self.cancelled:
            logger.debug('Scan was cancelled before it was started')
            return

        if num_points:
            self._progress_bar = tqdm.tqdm(total=num_points, unit='points')

        self._times['script_start'] = time.time()
        try:
            self._script_running = True
            ret = self.gpascii.run_and_wait(self.coord_sys, self.prog_num,
                                            verbose=False, stop_on_cancel=True,
                                            cancel_signal=self._cancelled,
                                            **kwargs)
            if ret != 0:
                logger.error('Script failed; ret code=%s', ret)
                self._failed = True
                return
        except ScriptCancelled as ex:
            logger.warning('Script cancelled (%s)', ex)
        except GPError as ex:
            logger.error('Script run error: (%s) %s', ex.__class__.__name__,
                         ex, )
            if 'READY TO RUN' in str(ex):
                logger.warning('Are all motors in the coordinate system in '
                               'closed loop?')
            raise
        finally:
            self._script_running = False

            if self.cancelled or self.failed:
                logger.debug('Running post-scan cancel script')
                cancel_script = os.path.join(self._script_path,
                                             self._cancel_script)
                try:
                    self.gpascii.run_simple_script(cancel_script)
                except Exception as ex:
                    logger.error('Failed to run cancel script %s',
                                 cancel_script, exc_info=ex)

            logger.debug('Script finished')
            self._times['script_finished'] = time.time()

            if self._shutter_control:
                if self.scan_count == self.scan_index:
                    logger.debug('Closing shutter')
                    shutter_close()

            if self.failed:
                pass
            elif self.cancelled:
                for det in self.detectors:
                    try:
                        det.stop()
                    except AttributeError:
                        pass

                logger.info('Scan %s partially complete. Processing data...',
                            self.scan_id_string)
            else:
                logger.info('Scan %s completed. Processing data...',
                            self.scan_id_string)

            data = None
            if not self.failed:
                data = self._process_data()

            logger.debug('Unstaging scan')
            try:
                self.unstage()
            except Exception as ex:
                logger.error('Unstaging failed', exc_info=ex)

        return data

    def run_subscan(self, index, **kwargs):
        '''Run a specific subscan'''
        # NOTE it's too late to update metadata here
        try:
            scan_info = self._subscans[index]
            self._scan_index = index + 1
            self._raw_gather_data = bytearray()

            macros = scan_info['macros']
            num_points = np.product(macros['points'])

            # move the axes to the start, while sending the motion program in
            # parallel
            self._move_all_axes(zip(macros['axis_names'],
                                    macros['scan_starts']),
                                wait=False)

            try:
                self._new_scan(**scan_info)
                # TODO make scripts part of the ppmac configuration and then
                #      only modify M/P/Q variables for parameters
                self.gpascii.send_program(self.coord_sys, self.prog_num,
                                          filename=scan_info['script_fn'],
                                          verbose=False, macros=macros,
                                          motors=self.number_to_axis)

                self._move_all_axes(zip(macros['axis_names'],
                                        macros['scan_starts']),
                                    wait=True,
                                    msg='moved to starting position')

                # Run the motion program
                return self._run_and_wait(variables=[self._sync_variable,
                                                     self._status_variable],
                                          change_callback=self._sync_callback,
                                          num_points=num_points,
                                          **kwargs)
            except Exception as ex:
                logger.error('Scan #%d of grid failed', self.scan_index,
                             exc_info=ex)
                raise
        finally:
            if self._progress_bar is not None:
                self._progress_bar.close()
                self._progress_bar = None

            last_scan = ((index + 1) == len(self._subscans))
            if self.cancelled or last_scan:
                self._move_all_axes(self._start_positions.items(),
                                    msg='moved back to',
                                    wait=False)



import IPython
import numpy as np
import logging
import time
from contextlib import contextmanager

import ophyd
import ophyd.flyers
# import ophyd.pseudopos
# from ophyd.device import Component as Cpt
from cycler import cycler
from .object_plans import Plan
from bluesky.run_engine import Msg
from bluesky.utils import Subs

from .callbacks.table import BulkLiveTable
from .log import log_time


logger = logging.getLogger(__name__)


@contextmanager
def elapsed_time_context(log_msg):
    t1 = time.time()
    yield
    t2 = time.time()

    t_dict = {'t1': t1,
              't2': t2}

    try:
        log_time(logger, t_dict, log_msg, t1='t1', t2='t2')
    except Exception:
        pass


def maybe_a_table(scan):
    ip = IPython.get_ipython()
    table_max_lines = ip.user_ns.get('table_max_lines', 5000)

    flyer = scan.flyer
    if flyer.total_points < table_max_lines:
        # logbook = ip.user_ns.get('logbook', None)
        # if logbook is not None:
        #   logbook = logbook.log
        return BulkLiveTable(fields=flyer.table_columns[:15]
                             # logbook=logbook,
                             )


def update_subs(subs, flyer):
    '''Update subscriptions with a flyer instance

    Only updates subscriptions with an existing 'flyer' attribute.
    '''
    for name, sub_list in subs.items():
        for sub in sub_list:
            if hasattr(sub, 'flyer'):
                sub.flyer = flyer


class FlyPlan(Plan):
    def __init__(self, usable_detectors=None, scaler_channels=None,
                 init_kw=None, *, md=None):
        super().__init__()

        if init_kw is None:
            init_kw = {}
        if scaler_channels is None:
            scaler_channels = [1, 2, 3, 4, 5, 6]
        if usable_detectors is None:
            ip = IPython.get_ipython()
            usable_detectors = [ip.user_ns[det_name] for det_name in
                                ('xspress3', 'merlin1', 'zebra', 'sclr1',
                                 'dexela1')
                                if det_name in ip.user_ns
                                ]

        self.usable_detectors = list(usable_detectors)
        self.scaler_channels = list(scaler_channels)
        self.init_kw = init_kw
        self.md = md if md is not None else {}
        self.motors = []

    def _instance_from_motors(self, *motors):
        self.init_kw.update(scaler_channels=self.scaler_channels)

        if not hasattr(self, 'scans') or not self.scans:
            raise ValueError('Scans must be setup')

        fset = frozenset(motors)
        try:
            cls = self.scans[fset]
        except KeyError:
            options = '\n'.join(', '.join(mtr.name for mtr in key)
                                for key in self.scans.keys())
            raise ValueError('Can only use these motors in a flyscan:' +
                             options)

        return cls(detectors=self.detectors, **self.init_kw)

    def _update_md(self):
        # update the metadata from various sources:
        for md in [self.flyer.md, self._configure_kw]:
            self.md.update(**md)

        # ensure we aren't trying to save devices in mds
        for key, value in self.md.items():
            if isinstance(value, (ophyd.Signal, ophyd.Device)):
                self.md[key] = value.name

        self.md['detectors'] = [det.name for det in self.detectors]
        self.md['motors'] = [motor.name for motor in
                             self.flyer._subscans[0]['positioners']]
        self.md['fast_axis'] = self.md['motors'][0]
        self.md['plan_name'] = self.__class__.__name__
        self.md['plan_type'] = self.__class__.__name__  # ?

        # remove macros that are in the top-level metadata to lessen redundancy
        if 'macros' not in self.md:
            return

        for macro, value in list(self.md['macros'].items()):
            if self.md.get(macro, None) == value:
                del self.md['macros'][macro]

    def __call__(self):
        # **TODO** calling configure twice to ensure subs get valid information
        #          on the flyer
        self.flyer.configure(**self._configure_kw)
        update_subs(self.subs, self.flyer)
        yield from super().__call__()

    def _gen(self):
        yield Msg('hxn_next_scan_id')
        yield Msg('configure', self.flyer, **self._configure_kw)
        self._update_md()
        yield Msg('open_run', **self.md)

        for subscan in range(self.flyer.scan_count):
            yield from flyer_plan(self.flyer, subscan=subscan)

        yield Msg('close_run')


def flyer_plan(flyer, subscan, *, name='primary'):
    '''Plan to run a flyer'''
    with elapsed_time_context('checkpoint'):
        yield Msg('checkpoint')
        # start the flyscan
        with elapsed_time_context('flyscan'):
            yield Msg('kickoff', flyer, subscan=subscan,
                      group='fly-kickoff', name=name)
            yield Msg('wait', None, 'fly-kickoff')

            # wait for the collect to be ready
            yield Msg('complete', flyer, group='fly-collect')
            yield Msg('wait', None, 'fly-collect')

        yield Msg('collect', flyer)


class FlyPlan1D(FlyPlan):
    '''1D fly scan

    A continuous, relative scan of `motor` from scan_start to scan_end,
    triggering detectors `scan_points` times.  The speed to move the positioner
    will be automatically determined to get a consistent `exposure_time`.

    Parameters
    ----------
    motor : Positioner
        The motor to scan
    scan_start : float
        Ending position (relative) [um]
    scan_end : float
        Ending position (relative) [um]
    scan_points : int
        Number of scan points
    exposure_time : float
        Exposure time in seconds

    dead_time : float, optional
        Between-frame dead time to allow detectors to write their data out.
        If set too low, some detectors may not acquire and scans may fail.
    return_speed : float
        The speed at which to return the motor to its starting position after
        the scan
    fly_type : str, optional
        'soft' uses trajectory generation and triggering via motion scripts
        'fly' normal 'fly-scan' concept, using position compare (note that this
        is not yet setup at the HXN)
    md : dict, optional
        Additional metadata to be saved
    '''
    sub_factories = Subs([maybe_a_table])

    def __call__(self, dets, motor, scan_start, scan_end, num, exposure_time, *,
                 dead_time=None, return_speed=None, fly_type=None, md=None):
        self.detectors = dets
        if md is not None:
            self.md.update(**md)

        self.flyer = self._instance_from_motors(motor)

        defaults = self.flyer.configure_defaults
        self._configure_kw = dict(
            motor=motor,
            scan_start=scan_start,
            scan_end=scan_end,
            num=num,
            exposure_time=exposure_time,
            dead_time=dead_time or defaults['dead_time'],
            return_speed=return_speed or defaults['return_speed'],
            fly_type=fly_type or defaults['fly_type']
        )

        self.motors = [motor]

        self.md['plan_args'] = dict(
            motor=repr(motor),
            start=scan_start, stop=scan_end, num=num,
            time=exposure_time,
            dead_time=self._configure_kw['dead_time'],
            return_speed=self._configure_kw['return_speed'],
            fly_type=self._configure_kw['fly_type'],
        )
        yield from super().__call__()


class FlyPlan2D(FlyPlan):
    '''2D fly scan

    A continuous, outer-product relative scan of
        `motor1` from scan_start1 to scan_end1 of num1 points
    and a relative scan of
        `motor2` from scan_start2 to scan_end2 of num2 points

    triggering detectors `num1` * `num2` times.  The speed to move the
    first positioner will be automatically determined to get a consistent
    `exposure_time`.

    Parameters
    ----------
    motor1 : Positioner
        The fast motor to scan
    scan_start1 : float
        Ending position (relative) [um]
    scan_end1 : float
        Ending position (relative) [um]
    num1 : int
        Number of scan points
    motor2 : Positioner
        The slow motor to scan
    scan_start2 : float
        Ending position (relative) [um]
    scan_end2 : float
        Ending position (relative) [um]
    num2 : int
        Number of scan points
    exposure_time : float
        Exposure time in seconds

    dead_time : float, optional
        Between-frame dead time to allow detectors to write their data out.
        If set too low, some detectors may not acquire and scans may fail.
    return_speed : float
        The speed at which to return the motor to its starting position after
        the scan
    fly_type : str, optional
        'soft' uses trajectory generation and triggering via motion scripts
        'fly' normal 'fly-scan' concept, using position compare (note that this
        is not yet setup at the HXN)
    max_points : int, optional
        Maximum number of scan points to execute in one 2D fly-scan (defaults
        to 16K, a limitation imposed by the Xspress3)
    md : dict, optional
        Additional metadata to be saved
    '''
    sub_factories = Subs([maybe_a_table])

    def __call__(self,
                 dets,
                 motor1, scan_start1, scan_end1, num1,
                 motor2, scan_start2, scan_end2, num2,
                 exposure_time,
                 *,
                 dead_time=None, fly_type=None,
                 return_speed=None, max_points=None, md=None):
        '''
        Additional user_kw arguments are passed to fly2d.Fly2D._configure

        Parameters not recognized by that method are sent to the metadatastore
        '''
        self.detectors = dets
        if md is not None:
            self.md.update(**md)

        self.flyer = self._instance_from_motors(motor1, motor2)

        defaults = self.flyer.configure_defaults
        self._configure_kw = dict(
            motor1=motor1,
            scan_start1=scan_start1,
            scan_end1=scan_end1,
            num1=num1,

            motor2=motor2,
            scan_start2=scan_start2,
            scan_end2=scan_end2,
            num2=num2,

            exposure_time=exposure_time,
            dead_time=dead_time or defaults['dead_time'],
            fly_type=fly_type or defaults['fly_type'],
            return_speed=return_speed or defaults['return_speed'],
            max_points=max_points or defaults['max_points'],
        )

        self.motors = [motor1, motor2]
        self.md['plan_args'] = dict(self._configure_kw)
        for key in ('motor1', 'motor2'):
            self.md['plan_args'][key] = repr(self.md['plan_args'][key])

        yield from super().__call__()


class FlyStep1D(FlyPlan):
    '''1D fly scan plus 1D step-scan

    A continuous, outer-product relative fly-scan of
        `motor1` from scan_start1 to scan_end1 of num1 points
    and a relative step-scan of
        `motor2` from scan_start2 to scan_end2 of num2 points

    triggering detectors `num1` * `num2` times.  The speed to move the
    first positioner will be automatically determined to get a consistent
    `exposure_time`.

    Parameters
    ----------
    motor1 : Positioner
        The fast motor to scan
    scan_start1 : float
        Ending position (relative) [um]
    scan_end1 : float
        Ending position (relative) [um]
    num1 : int
        Number of scan points
    motor2 : Positioner
        The slow motor to scan
    scan_start2 : float
        Ending position relative)
    scan_end2 : float
        Ending position (relative)
    num2 : int
        Number of scan points
    exposure_time : float
        Exposure time in seconds

    dead_time : float, optional
        Between-frame dead time to allow detectors to write their data out.
        If set too low, some detectors may not acquire and scans may fail.
    return_speed : float
        The speed at which to return the motor to its starting position after
        the scan
    fly_type : str, optional
        'soft' uses trajectory generation and triggering via motion scripts
        'fly' normal 'fly-scan' concept, using position compare (note that this
        is not yet setup at the HXN)
    md : dict, optional
        Additional metadata to be saved
    record_motor2 : bool, optional
        Record motor2 positions in a separate stream. Default is True.
        It may be preferable to disable this if all axes exist on the
        pmac controller
    '''
    sub_factories = Subs([maybe_a_table])

    def __call__(self,
                 dets,
                 motor1, scan_start1, scan_end1, num1,
                 motor2, scan_start2, scan_end2, num2,
                 exposure_time,
                 *,
                 dead_time=None, fly_type=None,
                 return_speed=None, md=None, record_motor2=True):
        '''
        Additional user_kw arguments are passed to fly2d.Fly2D._configure

        Parameters not recognized by that method are sent to the metadatastore
        '''
        self.detectors = dets
        if md is not None:
            self.md.update(**md)

        self.flyer = self._instance_from_motors(motor1)

        defaults = self.flyer.configure_defaults
        self._configure_kw = dict(
            motor=motor1,
            scan_start=scan_start1,
            scan_end=scan_end1,
            num=num1,

            exposure_time=exposure_time,
            dead_time=dead_time or defaults['dead_time'],
            fly_type=fly_type or defaults['fly_type'],
            return_speed=return_speed or defaults['return_speed'],
        )

        self.md['plan_args'] = dict(
            motor1=repr(motor1),
            scan_start1=scan_start1,
            scan_end1=scan_end1,
            num1=num1,

            motor2=repr(motor2),
            scan_start2=scan_start2,
            scan_end2=scan_end2,
            num2=num2,

            exposure_time=self._configure_kw['exposure_time'],
            dead_time=self._configure_kw['dead_time'],
            fly_type=self._configure_kw['fly_type'],
            return_speed=self._configure_kw['return_speed'],

            record_motor2=record_motor2,
        )

        self.motors = [motor1, motor2]

        # Record the steps the slow motor has to take:
        positions = np.linspace(motor2.position + scan_start2,
                                motor2.position + scan_end2,
                                num=num2, endpoint=True)
        self.slow_steps = cycler(motor2, positions)

        all_flyer_axes = all(key in self.flyer.positioner_to_axis for key in
                             self.slow_steps.keys)

        if not record_motor2 or all_flyer_axes:
            self.slow_flyer = None
        else:
            class SlowFlyer(ophyd.flyers.MonitorFlyerMixin, ophyd.Device):
                pass

            if isinstance(motor2, ophyd.EpicsMotor):
                motor2 = motor2.user_readback
            # elif isinstance(motor2, ophyd.pseudopos.PseudoPositioner):
            #     motor2 = motor2.user_readback

            self.slow_flyer = SlowFlyer('', name='slow_flyer', pivot=True)
            # TODO: this only supports one motor now, but is easy to extend
            self.slow_flyer.monitor_attrs = [motor2.name]
            self.slow_flyer.stream_names = {motor2.name: 'motor2'}
            setattr(self.slow_flyer, motor2.name, motor2)

            # for mtr in self.slow_steps.keys:
            #     if isinstance(mtr, ophyd.EpicsMotor):
            #         mtr = mtr.user_readback
            #     setattr(self.slow_flyer, mtr.name, mtr)

        yield from super().__call__()

    def _update_md(self):
        super()._update_md()

        motor2 = list(self.slow_steps.keys)[0]
        steps = self.slow_steps.by_key()[motor2]

        # metadata will be filled from the flyer, but we know a bit more
        # about this scan - it's actually 2d:
        plan_args = self.md['plan_args']
        self.md['shape'] = [plan_args['num1'], plan_args['num2']]
        self.md['motors'] = [self._configure_kw['motor'].name, motor2.name]
        self.md['scan_range'] = [self.md['scan_range'][0],
                                 [np.min(steps), np.max(steps)]]

    def _gen(self):
        # over-rides default gen
        yield Msg('hxn_next_scan_id')
        yield Msg('configure', self.flyer, **self._configure_kw)
        self._update_md()

        if len(self.flyer._subscans) != len(self.slow_steps):
            # fun hack: duplicate the sub-scan configuration for all
            # slow-steps in the flyer. this will allow liveimage to work
            # without modification.
            self.flyer._subscans = ([self.flyer._subscans[0]] *
                                    len(self.slow_steps))

        yield Msg('open_run', **self.md)

        if self.slow_flyer is not None:
            yield Msg('kickoff', self.slow_flyer, group='slowflyer-kickoff')
            yield Msg('wait', None, 'slowflyer-kickoff')

        for subscan, pt in enumerate(iter(self.slow_steps)):
            for mtr, pos in pt.items():
                yield Msg('set', mtr, pos, group='step-motors')
            yield Msg('wait', None, 'step-motors')
            yield from flyer_plan(self.flyer, subscan=subscan)

        if self.slow_flyer is not None:
            yield Msg('complete', self.slow_flyer, group='slowflyer-collect')
            yield Msg('wait', None, 'slowflyer-collect')
            yield Msg('collect', self.slow_flyer)

        yield Msg('close_run')



import asyncio
import logging
from collections import OrderedDict

import numpy as np
import matplotlib.pyplot as plt

from .flydata import (catch_exceptions, SubscanDataHandler)
from .liveplot import LivePlotBase
from .roiplot import FlyRoiPlot


loop = asyncio.get_event_loop()
logger = logging.getLogger(__name__)


def _sum_func(*values):
    return np.sum(values, axis=0)


def resize_1d(data, length):
    try:
        len(data)
    except TypeError:
        return np.zeros(length)

    if len(data) > length:
        return data[:length]
    elif len(data) < length:
        try:
            fill_value = np.nanmin(data)
        except ValueError:
            fill_value = 0.0

        _data = np.ones(length) * fill_value
        _data[:len(data)] = data
        return _data
    else:
        return data


def reshape(data, nx, ny, *, fly_type=None):
    data = resize_1d(data, nx * ny)
    data = data.reshape(ny, nx)

    if fly_type in ('pyramid', ):
        data[1::2, :] = data[1::2, ::-1]

    return data


class FlyLiveImage(FlyRoiPlot):
    # TODO this whole thing is a mess of inefficiency with array
    def __init__(self, roi_names, use_sum=False, data_func=None, **kwargs):
        super().__init__(roi_names, use_sum=use_sum, data_func=data_func,
                         **kwargs)

        self._gridspec = None

        if use_sum:
            data_func = _sum_func

        self.subscan_keys = OrderedDict((group_key,
                                         [sig.name for sig in roi_signals])
                                        for group_key, roi_signals in
                                        self.signals.items())

        logger.debug('Subscan keys %s', self.subscan_keys)
        self.subscan_dh = SubscanDataHandler(self.subscan_keys,
                                             data_func=data_func)
        self.subscan_data = OrderedDict()
        self.new_subscan = False

    def _subscan_recalculate(self):
        try:
            data = self.subscan_dh.calc_data(self.flyer, self.scan_id)
        except TypeError as ex:
            logger.debug('Subscan recalculation failed', exc_info=ex)
        else:
            self.subscan_data = data
            self.new_subscan = True

        return self.subscan_data

    def subscan_start(self, index=0, **kwargs):
        self._subscan_recalculate()

    def _reset(self):
        super()._reset()

    @catch_exceptions
    def _replot_preview(self):
        if self.new_subscan:
            self.new_subscan = False
        elif not self.data.updated:
            return

        try:
            nx, ny = self._run_header['shape']
        except ValueError:
            return

        try:
            npts = self.data.num_points
        except ValueError:
            npts = self.num

        npts = npts or 0

        subscan_data = self.subscan_data
        livedata = self.data.calc_data(npts)
        fly_type = self._run_header['fly_type']

        plot_data = OrderedDict()
        for group_key in self.subscan_keys.keys():
            arrs = [arr for arr in (subscan_data.get(group_key, []),
                                    livedata[group_key])
                    if np.size(arr)]

            if not arrs:
                continue
            elif len(arrs) == 1:
                data = arrs[0]
            else:
                data = np.concatenate(arrs)

            plot_data[group_key] = reshape(data, nx, ny,
                                           fly_type=fly_type)

        self.plot(self.fig, plot_data)
        self.fig.canvas.manager.set_window_title(f"LiveImage - scan {self.scan_id}")
        self.draw()

    def draw(self, fig=None):
        if fig is None:
            fig = self.fig

        fig.canvas.manager.show()
        fig.canvas.draw()
        fig.canvas.flush_events()

    @catch_exceptions
    def plot(self, fig, labeled_data, *, final=False, **kwargs):
        count = len(labeled_data)

        # The first frame may contain no data.
        if not count:
            return

        if final or self._gridspec is None or count != len(self._axes):
            fig.clear()
            rows = int(np.ceil(np.sqrt(count)))
            gridspec = plt.GridSpec(rows, rows)
            axes = OrderedDict((key, fig.add_subplot(gs))
                               for gs, key in
                               zip(gridspec, labeled_data.keys()))

            if not final:
                self._gridspec = gridspec
                self._axes = axes
        elif not final:
            axes = self._axes

        extent = kwargs.pop('extent', self.default_extents)
        for label, data in labeled_data.items():
            ax = axes[label]
            ax.clear()
            ax.imshow(data, interpolation='none', extent=extent, **kwargs)
            ax.set_title(label)
            ax.autoscale_view(tight=True)
            # self.cs.update_image(data)
        return axes

    def scan_started(self, doc, ndim, **scan_args):
        if ndim != 2:
            return

        self._gridspec = None  # Clear the figure at the next update
        self.subscan_data = OrderedDict()
        self.new_subscan = False

        (x1, x2), (y1, y2) = doc['scan_range']
        self.default_extents = [x1, x2, y2, y1]
        super().scan_started(doc, ndim, **scan_args)

    def get_motor_positions(self):
        motor1, motor2 = self._run_header['motors']
        subscan_data = self.flyer.data[self.scan_id]
        pos = [subscan_data[motor] for motor in (motor1, motor2)]
        return pos

    def get_plot_extents(self):
        try:
            pos = self.get_motor_positions()
        except KeyError:
            # Scan positions from mixed fly/step scans might not be available
            return self.default_extents
        except Exception as ex:
            logger.debug('Failed to get motor positions', exc_info=ex)
        else:
            mins = [np.min(p) for p in pos]
            maxes = [np.max(p) for p in pos]
            extent = [mins[0], maxes[0], maxes[1], mins[1]]

            if ((abs(extent[0] - extent[1]) <= 0.001) or
                    (abs(extent[2] - extent[3]) <= 0.001)):
                return None

            return extent

    @catch_exceptions
    def scan_finished(self, doc, scan_data, cancelled=False, **kwargs):
        try:
            nx, ny = self._run_header['shape']
            fly_type = self._run_header['fly_type']
        except (KeyError, ValueError):
            return

        subscan_data = self._subscan_recalculate()
        if not subscan_data:
            return

        # TODO poor oo-code :(
        LivePlotBase.scan_finished(self, doc, scan_data, cancelled=cancelled,
                                   **kwargs)

        extent = self.get_plot_extents()

        data = OrderedDict((label, reshape(d, nx, ny, fly_type=fly_type))
                           for label, d in subscan_data.items())

        # re-create the grid
        self._gridspec = None

        for fig in [self.fig, self.final_fig]:
            final = (fig is self.final_fig)
            self.plot(fig, data, final=final, extent=extent,
                      **self.plot_kwargs)
            self.draw(fig)
            fig.canvas.manager.set_window_title(f"Final image - scan {self.scan_id}")




import os
import time
import logging
import numpy as np

from .fly import FlyScan


logger = logging.getLogger(__name__)


class Fly2D(FlyScan):
    '''2D fly scan

    Scan types:
    * soft: uses trajectory generation and triggering via motion scripts
    * pyramid: same method as soft, but acquires in both forward and reverse
               directions
    * fly: normal 'fly-scan' concept, using position compare
    '''
    scripts = {'soft': 'soft_2d.txt',
               # 'fly': 'fly2d.txt',
               'pyramid': 'soft_2d_pyramid.txt',
               }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._subscans = []

    def subscan_config(self, motor1, scan_start1, scan_end1, num1,
                       motor2, scan_start2, scan_end2, num2, *,
                       exposure_time=None, dead_time=0.010, speed=None,
                       fly_type='soft', return_speed=10.0):
        """
        2D closed-loop fly-scan of ss(x,y,z), where motor1 is the fast motor
        """
        assert(motor1 in self.axes)
        assert(motor2 in self.axes)

        if num1 < 1 or num2 < 0:
            raise ValueError('Number of points should be >= 1')
        elif exposure_time < 0 or dead_time < 0:
            raise ValueError('Dwell time and dead time must be positive')

        axis_names = [motor1, motor2]
        points = np.array([num1, num2])

        scan_starts = np.array([scan_start1, scan_start2])
        scan_ends = np.array([scan_end1, scan_end2])
        final_pos = np.array(scan_ends)

        distance = np.abs(scan_ends - scan_starts)

        if exposure_time is not None:
            speeds = distance / (exposure_time * points)

        scan_starts = np.min([scan_starts, scan_ends], axis=0)
        scan_ends = np.max([scan_starts, scan_ends], axis=0)

        per_points = distance / points
        self.line_time = distance[0] / speeds[0]

        gpascii = self.gpascii
        sync_samples = gpascii.get_variable('Gather.MaxLines', type_=int)

        macros = {'axis_names': axis_names,
                  'coord_sys': self.coord_sys,
                  'dead_time': dead_time * 1000.0,  # expects milliseconds
                  'exposure_time': exposure_time,
                  'feed_rates': speeds,
                  'final_pos': final_pos,
                  'line_time': self.line_time * 1000.0,
                  'per_points': per_points,
                  'points': points,
                  'return_speed': return_speed,
                  'scan_ends': scan_ends,
                  'scan_starts': scan_starts,
                  'sync_samples': int(sync_samples / 2),
                  }

        script_fn = os.path.join(self._script_path, self.scripts[fly_type])
        return dict(dimensions=points.tolist(),
                    num_points=points[0] * points[1],
                    positioners=[self.axis_to_positioner[m]
                                 for m in (motor1, motor2)],
                    paths=[(scan_starts[i], scan_ends[i]) for i in (0, 1)],
                    exposure_time=exposure_time,
                    fly_type=fly_type,
                    script_fn=script_fn,
                    macros=macros,
                    )

    def _configure(self, motor1, scan_start1, scan_end1, num1,
                   motor2, scan_start2, scan_end2, num2, *,
                   dead_time=0.007, relative=True, exposure_time=0.1,
                   max_points=None, fly_type='soft', return_speed=10.0):

        super()._configure()

        name1 = self._get_axis_name(motor1)
        name2 = self._get_axis_name(motor2)
        scan_start1, scan_end1 = float(scan_start1), float(scan_end1)
        scan_start2, scan_end2 = float(scan_start2), float(scan_end2)
        num1, num2 = int(num1), int(num2)
        exposure_time = float(exposure_time)
        dead_time = float(dead_time)

        if relative:
            start_pos = [self._start_positions[m]
                         for m in (name1, name2)]
        else:
            start_pos = [0.0, 0.0]

        points = _get_scan_points(scan_start1, scan_end1, num1,
                                  scan_start2, scan_end2, num2,
                                  max_points=max_points,
                                  start_x=start_pos[0],
                                  start_y=start_pos[1])
        points = list(points)

        self._subscans = [self.subscan_config(name1, _x1, _x2, nx,
                                              name2, _y1, _y2, ny,
                                              dead_time=dead_time,
                                              exposure_time=exposure_time,
                                              speed=None, fly_type=fly_type,
                                              return_speed=return_speed)
                          for i, (_x1, _x2, nx, _y1, _y2, ny)
                          in enumerate(points)]

        self._metadata['subscan_dims'] = [info['dimensions']
                                          for info in self._subscans]

        self.estimated_time = 0.0
        dim2 = int(np.sum([_ny for _nx, _ny in
                           self._metadata['subscan_dims']]))

        self._metadata['shape'] = [num1, dim2]
        self._total_points = num1 * dim2

        self._metadata['scan_count'] = self.scan_count
        self._metadata['scan_range'] = [(start_pos[0] + scan_start1,
                                         start_pos[0] + scan_end1),
                                        (start_pos[1] + scan_start2,
                                         start_pos[1] + scan_end2)]

        if self._subscans:
            line_time = self._subscans[0]['macros']['line_time']
            self.estimated_time = line_time / 1000.0 * num2

            for key in ('scan_starts', 'scan_ends', 'final_pos', 'per_points',
                        'points'):
                values = np.asarray([subscan['macros'][key] for subscan in
                                     self._subscans])
                self._metadata['{}'.format(key)] = values.tolist()

        return self._subscans


def _get_scan_points(x1, x2, total_pointsx,
                     y1, y2, total_pointsy, *,
                     max_points=None,
                     start_x=0., start_y=0.):
    '''
    Generates rectangular strips for a given scan range

    Maximum points, by default, is set to 16,384 for the xspress3
    '''
    # TODO: not so much limited by the number of points anymore
    #       - up to 16k frames for the xspress3
    #       however, need to break it up based on the travel range
    #       of the fine/coarse piezos
    if max_points is None:
        max_points = 16384

    assert x2 > x1
    assert y2 > y1

    dy = abs(y2 - y1) / total_pointsy

    single_scan_num_lines = max_points // total_pointsx
    if single_scan_num_lines <= 1:
        raise ValueError("One scan line won't fit with maximum points")

    # ensure the number of lines is even to ease data analysis with snaked/
    # pyramid scans
    if (single_scan_num_lines % 2) == 1:
        single_scan_num_lines -= 1

    pyi = 0

    xi = start_x + x1
    xj = start_x + x2

    while pyi < total_pointsy:
        remain = total_pointsy - pyi
        if remain < single_scan_num_lines:
            pointsy = remain
        else:
            pointsy = single_scan_num_lines

        yi = start_y + y1 + pyi * dy
        yj = yi + pointsy * dy

        yield xi, xj, total_pointsx, yi, yj, pointsy

        pyi += pointsy



import time

import pytest
import numpy as np
import pandas as pd

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

import filestore
import filestore.api
import filestore.handlers

import hxnfly.fly
import hxnfly.log
from hxnfly.callbacks import FlyLiveCrossSection
from hxnfly.fly2d import Fly2D
from hxnfly.bs import FlyPlan2D

from .fly_mock import (MockDetector, MockSignal)
from .sim_detector import TestDetector
from .fixtures import *


def test_scan_points():
    from hxnfly.fly2d import _get_scan_points
    point_args = (0.0, 1.0, 10,
                  1.0, 2.0, 10,
                  )
    points = list(_get_scan_points(*point_args, max_points=22))

    assert len(points) == 5
    assert sum(py for xi, xj, px, yi, yj, py in points) == 10

    xis = [xi for xi, xj, px, yi, yj, py in points]
    xjs = [xj for xi, xj, px, yi, yj, py in points]
    assert min(xis) == 0.0
    assert max(xjs) == 1.0

    yis = [yi for xi, xj, px, yi, yj, py in points]
    yjs = [yj for xi, xj, px, yi, yj, py in points]
    assert min(yis) == 1.0
    assert max(yjs) == 2.0


def make_2d_data(startx, endx, nptsx,
                 starty, endy, nptsy,
                 gathered_points=200,
                 gate_enable='low_to_high'):
    # this is not entirely accurate...
    time = np.linspace(0, 10.0, gathered_points)
    gather_y = gathered_points // nptsy
    px = np.array(np.linspace(startx, endx, gather_y).tolist() * nptsy)
    py = np.array(sum(([pt] * gather_y
                       for pt in np.linspace(starty, endy, nptsy)),
                      []))
    pz = np.random.rand(gathered_points) / 10.0

    enable = []
    npts = nptsx * nptsy
    exposed_count = (gathered_points // npts) - 1

    for i in range(npts):
        if gate_enable == 'low_to_high':
            enable.extend([0] + [1] * exposed_count)
        else:
            enable.extend([1] + [0] * exposed_count)

    if gate_enable == 'low_to_high':
        enable[-1] = 0
    else:
        enable[-1] = 1

    return [np.array(v) for v in (time, enable, px, py, pz)]


@pytest.fixture(scope='function',
                params=['with_mock_detector', 'with_sim_detector',
                        'relative'])
def fly2d(request, monkeypatch, ppmac, gpascii, axes, positioners,
          sim_det, ipython, global_state):
    import hxntools.scans
    hxntools.scans.setup(debug_mode=True)

    def run_and_wait(*args, **kwargs):
        print('run and wait!')
        change_callback = kwargs.pop('change_callback')
        time.sleep(0.1)
        change_callback('Q1', 0, 1)
        time.sleep(1.0)
        return 0

    monkeypatch.setattr(gpascii, 'run_and_wait', run_and_wait)

    if request.param == 'with_sim_detector':
        sim_det.count_time.put(0.01)
        detectors = [sim_det]
    else:
        detectors = [MockDetector('', name='det')]

    gpascii.set_variable('gather.samples', 100)
    gpascii.set_variable('gather.maxlines', 100)

    scan = Fly2D(axes=axes, positioners=positioners, detectors=detectors)
    startx, endx, nptsx = -1.0, 1.0, 2
    starty, endy, nptsy = -1.0, 1.0, 2
    exposure_time = 1.0
    relative = (request.param == 'relative')
    scan.configure(positioners['testx'], startx, endx, nptsx,
                   positioners['testy'], starty, endy, nptsy,
                   exposure_time=exposure_time, relative=relative)

    npts = nptsx * nptsy
    sclr1 = ipython.user_ns['sclr1']
    sclr1.mca_by_index[1].spectrum.put(np.array([exposure_time * 50e3] * npts))

    gather_client = scan.gather_client
    # fake data for parsing
    gather_client.gathered = make_2d_data(startx, endx, nptsx,
                                          starty, endy, nptsy,
                                          gathered_points=500,
                                          gate_enable=scan._gate_enable)

    def FakeClass(*args, **kwargs):
        print('initialized with', args, kwargs)
        scan.configure_defaults = dict(return_speed=5.0, dead_time=0.007,
                                       fly_type='soft', max_points=16384)
        return scan

    testx, testy = positioners['testx'], positioners['testy']
    FlyPlan2D.scans = {frozenset({testx, testy}): FakeClass,
                       }
    return scan


def test_fly2d(fly2d):
    # kickoff sends in the subscan number normally
    data = fly2d.run_subscan(0)
    df = pd.DataFrame(list(pd.DataFrame(data)['data']))
    print('acquired data:')
    print(df)

    has_test_det = any(isinstance(det, TestDetector)
                       for det in fly2d.detectors)

    if has_test_det:
        assert 'sim_tiff' in df

        for img_uid in df['sim_tiff']:
            print(img_uid, filestore.api.retrieve(img_uid).shape)

    print(fly2d.collect())
    print(fly2d.describe())


def test_failed_fly2d(fly2d, gpascii, monkeypatch):
    def run_and_wait(*args, **kwargs):
        time.sleep(1.0)
        return 1

    monkeypatch.setattr(gpascii, 'run_and_wait', run_and_wait)
    data = fly2d.run_subscan(0)
    assert data is None


def test_flyplan2d(monkeypatch, positioners, fly2d, run_engine):
    scan_fcn = hxnfly.bs.FlyPlan2D()
    gen = scan_fcn(positioners['testx'], -1.0, 1.0, 2,
                   positioners['testy'], -1.0, 1.0, 2,
                   1.0, dead_time=0.001,
                   )
    run_engine(gen)


def test_flyplan2d_liveimage(request, monkeypatch, positioners, run_engine,
                             fly2d, xspress3):
    fly2d.detectors.append(xspress3)

    scan_fcn = hxnfly.bs.FlyPlan2D()
    from hxnfly.callbacks import (FlyDataCallbacks, FlyLiveImage)

    monkeypatch.setattr(hxnfly.fly, 'Xspress3Detector', xspress3.__class__)

    xspress3.array_counter.put(4)

    # TODO is this done in configure_roi?
    xspress3.setup_fake_rois([('Fe', [1, 2, 3, 4]),
                              ('Mo', [4, 3, 2, 1])
                              ])
    liveimage = FlyLiveImage(['Fe', 'Mo'])

    scan_fcn.subs = [FlyDataCallbacks(), liveimage]
    gen = scan_fcn(positioners['testx'], -1.0, 1.0, 2,
                   positioners['testy'], -1.0, 1.0, 2,
                   1.0, dead_time=0.001,
                   )

    run_engine(gen)

    plt.savefig('liveimage-{}.png'.format(request.node.name))
    liveimage.disable()
    plt.clf()


# TODO crossection plot needs fixing
# @pytest.mark=
def test_flyplan2d_crossection(request, monkeypatch, positioners, run_engine,
                               fly2d, xspress3):
    fly2d.detectors.append(xspress3)

    scan_fcn = hxnfly.bs.FlyPlan2D()

    monkeypatch.setattr(hxnfly.fly, 'Xspress3Detector', xspress3.__class__)

    xspress3.array_counter.put(4)

    # TODO is this done in configure_roi?
    xspress3.setup_fake_rois([('Fe', [1, 2, 3, 4]),
                              ('Mo', [4, 3, 2, 1])
                              ])
    with pytest.raises(ValueError):
        # cross-section only does 1 at a time
        FlyLiveCrossSection(['Fe', 'Mo'])

    crossection = FlyLiveCrossSection(['Fe'])
    scan_fcn.subs = [crossection]
    gen = scan_fcn(positioners['testx'], -1.0, 1.0, 2,
                   positioners['testy'], -1.0, 1.0, 2,
                   1.0, dead_time=0.001,
                   )

    run_engine(gen)

    crossection.disable()
    from PyQt4.QtGui import QPixmap

    live_fn = 'crossection-live-{}.png'.format(request.node.name)
    QPixmap.grabWindow(crossection.live_window.winId()).save(live_fn, 'png')
    crossection.live_window.close()

    if crossection._final_window is not None:
        final_fn = 'crossection-final-{}.png'.format(request.node.name)
        QPixmap.grabWindow(crossection.final_window.winId()).save(final_fn,
                                                                  'png')
        crossection.final_window.close()



import os
import logging

from .fly import FlyScan


logger = logging.getLogger(__name__)


class Fly1D(FlyScan):
    '''1D fly scan

    Scan types:
    * soft: uses trajectory generation and triggering via motion scripts
    * fly: normal 'fly-scan' concept, using position compare
    '''
    scripts = {'soft': 'soft_1d.txt',
               'fly': 'fly1d.txt',
               }

    def subscan_config(self, motor, scan_start, scan_end, scan_points,
                       exposure_time=None, *, dead_time=0.010, speed=None,
                       fly_type='soft', relative=True, return_speed=0.4):

        axis = self._get_axis_name(motor)
        scan_start, scan_end = float(scan_start), float(scan_end)
        exposure_time = float(exposure_time)
        dead_time = float(dead_time)
        scan_points = int(scan_points)

        assert(axis in self.axes)

        if scan_points <= 1:
            raise ValueError('Number of points should be > 1')
        elif exposure_time < 0 or dead_time < 0:
            raise ValueError('Dwell time and dead time must be positive')

        if relative:
            all_pos = self.positions
            abs_pos = all_pos[axis]
            final_pos = abs_pos

            scan_start = abs_pos + scan_start
            scan_end = abs_pos + scan_end
        else:
            final_pos = scan_end

        distance = abs(scan_end - scan_start)

        if exposure_time is not None:
            speed = distance / (exposure_time * scan_points)

        per_point = distance / scan_points
        scan_start = min(scan_start, scan_end)
        scan_end = max(scan_start, scan_end)
        self.line_time = distance / speed

        macros = {'axis_name': axis,
                  'axis_names': [axis],
                  'coord_sys': self.coord_sys,
                  'dead_time': dead_time * 1000.0,  # expects milliseconds
                  'exposure_time': exposure_time,
                  'feed_rate': speed,
                  'final_pos': final_pos,
                  'line_time': self.line_time * 1000.0,
                  'per_point': per_point,
                  'points': scan_points,
                  'return_speed': return_speed,
                  'scan_start': scan_start,
                  'scan_starts': [scan_start],
                  }

        self._metadata['macros'] = {}
        for macro, value in sorted(macros.items()):
            logger.debug('Macro %s: %s', macro, value)
            self._metadata['macros'][macro] = value

        self._metadata['shape'] = [scan_points]
        self._metadata['scan_range'] = [[scan_start, scan_end]]
        self._total_points = scan_points
        self.estimated_time = self.line_time

        return dict(num_points=scan_points,
                    dimensions=[scan_points],
                    positioners=[self.axis_to_positioner[axis]],
                    paths=[(scan_start, scan_end)],
                    exposure_time=exposure_time,
                    fly_type=fly_type,
                    macros=macros,
                    script_fn=os.path.join(self._script_path,
                                           self.scripts[fly_type]),
                    )

    def _configure(self, motor, scan_start, scan_end, num,
                   exposure_time=None, *, dead_time=0.010, speed=None,
                   fly_type='soft', relative=True, return_speed=0.4):
        """
        1D closed-loop fly-scan of ss(x,y,z)
        """
        super()._configure()
        conf = self.subscan_config(motor, scan_start, scan_end, num,
                                   exposure_time=exposure_time,
                                   dead_time=dead_time,
                                   speed=speed,
                                   fly_type=fly_type,
                                   relative=relative,
                                   return_speed=return_speed)

        self._subscans = [conf]
        return self._subscans



import asyncio
import logging
from collections import OrderedDict

import PyQt5
import IPython

import matplotlib.pyplot as plt

from .flydata import (FlyDataCallbacks, catch_exceptions, SignalDataHandler)

from matplotlib.backends.qt_compat import QtCore

# loop = asyncio.get_event_loop()
logger = logging.getLogger(__name__)


def get_insertFig():
    ip = IPython.get_ipython()
    try:
        return ip.user_ns['insertFig']
    except KeyError:
        def insertFig(*args, **kwargs):
            logger.error('insertFig not found in user namespace')
        return insertFig


def add_toolbar_button(fig, text, *, theme_icon='go-down', slot=None):
    toolbar = fig.canvas.manager.toolbar
    # QtWidgets.QToolBar
    icon = PyQt5.QtGui.QIcon.fromTheme(theme_icon)
    action = toolbar.addAction(icon, text)
    if slot is not None:
        action.triggered.connect(slot)
    return action


class LivePlotBase(FlyDataCallbacks):
    def __init__(self, signals, flyer=None, rate=2.0, point_signal=None,
                 data_func=None, enabled=True,
                 **plot_kwargs):
        super().__init__(flyer=flyer)

        if not signals:
            raise ValueError('Must have at least one signal to plot')

        self._continue_updates = False

        self.signals = signals
        self.data = SignalDataHandler(signals, data_func=data_func,
                                      point_signal=point_signal)

        self.plot_kwargs = plot_kwargs
        self.rate = rate
        self.enabled = enabled

        self.final_ax = None
        self.final_fig = None
        self.fig, self.ax = None, None

    def _reset(self):
        if self.fig is None or not plt.fignum_exists(self.fig.number):
            self.fig, self.ax = plt.subplots()
            add_toolbar_button(self.fig, 'Call insertFig()',
                               slot=lambda fig: get_insertFig()(fig=self.fig))
        else:
            self.fig.clear()
            self.ax = self.fig.add_subplot(111)

    def disable(self):
        '''Turn off plotting'''
        self.enabled = False

    def enable(self):
        '''Turn on plotting'''
        self.enabled = True

    def reset(self):
        '''Open new windows on the next scan'''
        # self.fast_axis = None
        pass

    def close_all(self):
        '''Close all plots'''
        plt.close('all')

    @catch_exceptions
    def _update(self):
        try:
            if self._run_header is not None:
                self._replot_preview()
        finally:
            # loop.call_later(self.rate, self._update)
            if self._continue_updates:
                QtCore.QTimer.singleShot(self.rate * 1000, self._update)


    def scan_started(self, doc, ndim, fast_axis=None, **scan_args):
        # if self.fast_axis != fast_axis:
        self._reset()

    def scan_finished(self, doc, scan_data, cancelled=False, **kwargs):
        self._stop_updates()

        if (self.final_fig is None or not
                plt.fignum_exists(self.final_fig.number)):
            fig, ax = plt.subplots()
            self.final_ax = ax
            self.final_fig = fig
            self.final_legend = None
            add_toolbar_button(self.final_fig, 'Call insertFig()',
                               slot=lambda fig: get_insertFig()(fig=self.final_fig))

    def _start_updates(self):
        # loop.call_soon(self._update)
        self._continue_updates = True
        QtCore.QTimer.singleShot(1000, self._update)  # First update after 1s

    def _stop_updates(self):
        self._continue_updates = False


class FlyLivePlot(LivePlotBase):
    def __init__(self, signals, y='counts', yscale='linear', **kwargs):
        super().__init__(signals, **kwargs)

        self.yscale = yscale
        self.ylabel = y
        self.enabled = True

    def _reset(self):
        super()._reset()

        self.lines = []
        self.preview_lines = None

        self.ax.set_ylabel(self.ylabel or 'counts')
        self.ax.set_xlabel('Scan point')

        self.ax.set_yscale(self.yscale)
        self.ax.margins(0.1)

    @catch_exceptions
    def _replot_preview(self):
        try:
            npts = self.data.num_points
        except ValueError:
            npts = self.num

        npts = npts or 0

        with self.data._lock:
            if not self.data.updated:
                return

            data = self.data.calc_data(npts)

        if self.preview_lines is None:
            def _plot(label, linedata):
                return self.ax.plot(range(len(linedata)), linedata,
                                    label=label, **self.plot_kwargs)[0]

            lines = OrderedDict((label, _plot(label, linedata))
                                for label, linedata in data.items())
            self.preview_lines = lines
            self.legend = self.ax.legend(loc=0)
            if self.legend is not None:
                self.legend.set_draggable(True)
        else:
            for key, line in self.preview_lines.items():
                linedata = data[key]
                xpts = range(len(linedata))
                line.set_data(xpts, linedata)

        # Rescale and redraw.
        self.ax.relim(visible_only=True)
        self.ax.autoscale_view(tight=True)
        self.fig.canvas.draw()

    def scan_started(self, doc, ndim, **scan_args):
        super().scan_started(doc, ndim, **scan_args)

        self.ax.set_title('Scan {}'.format(self.scan_id))
        self._start_updates()

    def scan_finished(self, doc, scan_data, cancelled=False, **kwargs):
        super().scan_finished(doc, scan_data, cancelled=cancelled, **kwargs)

        ax, fig = self.final_ax, self.final_fig

        if scan_data is None:
            logger.error('No scan data for the live plot')
            return

        try:
            x_pos = scan_data[self.fast_axis]
        except KeyError as ex:
            logger.error('Motor position data not found in flyscan '
                         'instance?', exc_info=ex)
            return

        npts = len(x_pos)
        data = self.data.calc_data(npts)
        for label, y_data in data.items():
            label = 'S{} {}'.format(self.scan_id, label)
            try:
                length = min((npts, len(y_data)))
                ax.plot(x_pos[:length], y_data[:length], label=label,
                        **self.plot_kwargs)
            except TypeError:
                continue

        self.final_legend = ax.legend(loc=0)
        if self.final_legend is not None:
            self.final_legend.set_draggable(True)

        ax.set_ylabel(self.ylabel or 'counts')
        ax.set_xlabel(self.fast_axis)

        # Rescale and redraw.
        ax.set_yscale(self.yscale)
        ax.relim(visible_only=True)
        ax.autoscale_view(tight=True)
        fig.canvas.draw()



import logging
from collections import OrderedDict
from IPython import get_ipython

from .fly1d import Fly1D
from .fly2d import Fly2D

logger = logging.getLogger(__name__)

ip = get_ipython()

try:
    dssx = ip.user_ns['dssx']
    dssy = ip.user_ns['dssy']
    dssz = ip.user_ns['dssz']
except KeyError as ex:
    logger.error('Ensure that fly scan motor "%s" is still in the startup '
                 'configuration with the same name.', ex,
                 exc_info=ex)
    raise

coord_sys = 7
prog_num = 50
# swap x and z axes: interchange axis_numbers of x and z (H. Yan, 01/10/17)
# physically swapped cable for dssx and dssz; change the seeting to their original state (H. Yan, 08/08/2019)
axes = OrderedDict([('x', dict(positioner=dssx, axis_number=9)),
                    ('y', dict(positioner=dssy, axis_number=10)),
                    ('z', dict(positioner=dssz, axis_number=11))]
                    )
# change return_speed from 5.0 to 40 (H. Yan, 01/10/17)
configure_defaults = dict(return_speed=40.0,
                          dead_time=0.002,
                          fly_type='soft',
                          max_points=16384
                          )


class Fly1D_Diffraction(Fly1D):
    def __init__(self, cmd='fly1d', **kwargs):
        super().__init__(axes=axes, coord_sys=coord_sys, prog_num=prog_num,
                         cmd=cmd, configure_defaults=configure_defaults,
                         **kwargs)


class Fly2D_Diffraction(Fly2D):
    def __init__(self, cmd='fly2d', **kwargs):
        super().__init__(axes=axes, coord_sys=coord_sys, prog_num=prog_num,
                         cmd=cmd, configure_defaults=configure_defaults,
                         **kwargs)



from __future__ import print_function
import versioneer

try:
    from setuptools import setup, find_packages
except ImportError:
    try:
        from setuptools.core import setup
    except ImportError:
        from distutils.core import setup


setup(name='hxnfly',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      author='HXN',
      packages=['hxnfly', 'hxnfly.callbacks'],
      package_data={'hxnfly': ['scripts/*.txt']})



import numpy as np
import matplotlib.pyplot as plt


def plot_position(scan_id, data, debug_data, *, left='ssx', right='ssy'):
    start_t = debug_data['start_time']
    end_t = debug_data['end_time']
    t = debug_data['servo_time']
    gather_x = debug_data['servo_' + left]
    gather_y = debug_data['servo_' + right]
    target_x = debug_data.get('target_' + left, None)
    target_y = debug_data.get('target_' + right, None)
    avg_x = data[left]
    avg_y = data[right]
    # gather_z = data['servo_ssz']
    mid_t = data['elapsed_time']

    colors = {'x': 'blue',
              'y': 'red',
              'z': 'green',
              'target_x': 'black',
              'target_y': 'black',
              'target_z': 'black'
              }

    plt.clf()
    y_min, y_max = np.min(gather_x), np.max(gather_x)

    # frame start/end times (green = start, red = end)
    plt.vlines(start_t, y_min, y_max, colors='g', linestyle='solid')
    plt.vlines(end_t, y_min, y_max, colors='r', linestyle='dashed')

    plt.ylabel('x position [um]')
    plt.plot(t, gather_x, colors['x'], label='fssx')
    if target_x is not None:
        plt.plot(t, target_x, colors['target_x'], label='target_fssx')
    plt.scatter(mid_t, avg_x, marker='o', color=colors['x'],
                zorder=10)

    ax1 = plt.gca()
    ax1.yaxis.label.set_color(colors['x'])
    ax1.tick_params(axis='y', colors=colors['x'])

    ax2 = plt.gca().twinx()
    plt.ylabel('y position [um]')
    ax2.plot(t, gather_y, colors['y'], label='fssy')
    if target_y is not None:
        plt.plot(t, target_y, colors['target_y'], label='target_fssy')

    ax2.yaxis.label.set_color(colors['y'])
    ax2.tick_params(axis='y', colors=colors['y'])

    y_min = np.min(gather_y)
    y_max = np.max(gather_y)

    # frame start/end times (green = start, red = end)
    ax2.vlines(start_t, y_min, y_max, colors='g', linestyle='solid')
    ax2.vlines(end_t, y_min, y_max, colors='r', linestyle='dashed')

    ax2.scatter(mid_t, avg_y, marker='o', color=colors['y'],
                zorder=10)

    plt.xlim(0, t[-1])
    plt.title('Scan {} positions'.format(scan_id, ))
    plt.xlabel('Time [s]')
    return [ax1, ax2]
